#!/usr/bin/env python
# -*- coding: utf-8 -*-



import os, sys, gtk, random

############################################################################################################################################################
########################################################## START gimpfu3 by J.F.Garcia [arakne] ############################################################
############################################################################################################################################################
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb

import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}

def register(proc_name, blurb, help, author, copyright, date, label, imagetypes, params, results, function, menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
		return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead" % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]], x[1], _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode", "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results)
        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
    gimpui.gimp_ui_init ()  # add MrQ	
#    import pango

    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default=""):
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            #self.set_line_wrap(True)
            #self.set_single_line_mode(False)

        def get_value(self):
            return self.get_text()

    class LabelTitle(gtk.Label):
        def __init__(self, default="", align = (0.0, 0.0)):
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default=""):
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))

    class ToggleEntry(gtk.ToggleButton):
        def __init__(self, default=0, labels=[_("No"),_("Yes")]):
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()
            self.connect("toggled", self.changed)
            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0))):
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=()):
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)

        def get_value(self):
            return self.get_active()

    def FileSelector(default=""):
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            selector = DirnameSelector
            if default == "/": default = ""
        else:
            selector = FilenameSelector
        return selector(default)

    class FilenameSelector(gtk.FileChooserButton):
        def __init__(self, default="", save_mode=False):
            gtk.FileChooserButton.__init__(self,_("Python-Fu File Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default=""):
            gtk.FileChooserButton.__init__(self,_("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : gimpui.ColorSelector,
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : gimpui.FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : gimpui.PatternSelector,
            PF_GRADIENT    : gimpui.GradientSelector,
            PF_PALETTE     : gimpui.PaletteSelector,
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    vbox = gtk.VBox(False, 12)
    vbox.set_border_width(12)
    dialog.vbox.pack_start(vbox)
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})

    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()

    iniX=0
    dialog.connect("response", response)
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]
 
        tab, incr, rows, connect, cols, hlp = (0, 1, 1, None, 1, desc)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            hlp  = getXtra(xtra, 'help'   , desc)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE and desc != '':
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW ):
                wid = _edit_mapping[pf_type](def_val, paramsi[4])
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4])
                else:
                    wid = _edit_mapping[pf_type](def_val)
            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val)
            if pf_type!=PF_TITLE and desc != '':
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', hlp)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()

    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break
        min_args = i
    if len(script_params) > min_args:
        start_params = params[:min_args + 1]
        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE
    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])
        gimp.displays_flush()
    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message




############################################################################################################################################################
######################################################### END gimpfu3 ######################################################################################
############################################################################################################################################################

#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2023 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
# It was written after receiving a request from Pat625 on GimpLearn
# It can be executed by selecting the menu option: Filters>Custom Fonts>Using Tools/Custom Font Paragrapher vx.xx Tabbed...
#
# The filter uses the function "computeThetaWithSlope" written by Ofnuts
# and parts of code taken from Tranductin's filter, custom-font_on_path, which he attributes to a post on http://stackoverflow.com/questions/31735499/calculate-angle-clockwise-between-two-points
# Without their work I would not have been able to create this filter
#



#from gtk import gdk
#w = gdk.screen_width()
#h = gdk.screen_height()

# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)


def do_stuff(image, layer, xcffilename, text, font_size, line_spacing, kern_amount, indent_type, indent_size, # 1st tab
            top_margin, left_margin, right_margin, bottom_margin, merge_options, user_path, follow_path, remove_user_path, # 2nd tab
            background_type, col_style, gradient_style, selection_width, feather_val, do_resize,  # 3rd tab
            drop_colour, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, ptab): # 4th tab
			
    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)
	
    dummy = 0
	
    
    text = text.decode('utf-8')
    dummy = 0
    custom_font_test, fontimg = fn_test_valid_custom_font(xcffilename) # check that the user has selected a valid custom font
	
	
    if custom_font_test:
        im_resize = False
		
		
		
        indent_width = 0
        if indent_type==1: indent_width = indent_size
        if text == "": text = "Emergency text used for now. Enter some of your own text next time."
        text = fn_check_text_is_valid(text, fontimg) # check that characters exist and swap case if necessary
		
        text = text.strip() # remove any leading or trailing space characters
		
		
			
        checked_text, prev_char = "", ""
        loop = 0
        while loop<len(text)-1: # remove space before newline event from user text
            char, next_char = text[loop], text[loop+1]
            if char==" " and next_char=="\n":
                dummy=0
            else:
                checked_text += char
            #    
				
            loop+=1
        checked_text+=text[loop]
        text = checked_text
        
		
		
        checked_text, prev_char = "", ""
        for char in text:
            #    
            # remove multiple space characters from the user text
            if (char == " " and prev_char == " "):
                dummy = 0
            else:
                checked_text += char
				
            prev_char = char
        text = checked_text
		
		
        # find out if the text contains spaces
        result = text.find(" ")
        if result >=0:
            no_space = False
        else:
            no_space = True
			
		
        # get width and height to use for space character for the current font size
        space_name = "space.png"
        char_name = "space"
        layer_name = char_name + ".png"
        char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
        custom_scale = fn_get_custom_font_scale(font_size, char_layer.height)
        if char_layer < 0: #layer not found
            space_width = (font_size/2)*custom_scale # Make up a space width for this font size
            space_height = font_size*custom_scale
            custom_height = space_height
            custom_width = space_width
        else:
            custom_scale = fn_get_custom_font_scale(font_size, char_layer.height)
            space_width = char_layer.width * custom_scale
            space_height = char_layer.height * custom_scale
            custom_height = space_height
            custom_width = space_width
			
        indent_width = indent_width*space_width
			
			
			
        font_y_offset = (space_height * 1.164)/2 # check this
		
        if user_path: # check that the user has provided a path to use with this option
            
            num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
            vector_list = image.vectors
            if num_vectors == 0: user_path = False # revert to straight line path if there is no user path			
			
        if user_path:
            # Get the current active path i.e. the newly created temp path
            new_vector_id = pdb.gimp_image_get_active_vectors(image)
            if new_vector_id == -1:
                pdb.gimp_image_set_active_vectors(image, vector_ids[0])# use the first path for the active path
                new_vector_id = pdb.gimp_image_get_active_vectors(image)
                new_vector_id_name = pdb.gimp_item_get_name(new_vector_id)
            new_vector_id_name = pdb.gimp_item_get_name(new_vector_id)
            # Get path stroke information
            NumberOfStrokes_in_Path,new_vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(new_vector_id)
            new_vector_StrokeID = new_vector_StrokeIDs_List[0]
			
            line_path_width = pdb.gimp_vectors_stroke_get_length(new_vector_id,new_vector_StrokeID,0.01) # get width of the path
			
            # find smallest x and y points along user path
            dist = 0
            user_path_min_x, user_path_min_y = 0, 0
            while dist<=line_path_width:
                x_point,y_point,SlopeAtThisPoint,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(new_vector_id,new_vector_StrokeID,dist,0.01)
                if dist == 0:
                      user_path_min_x = x_point
                      user_path_min_y = y_point
                else:
				
                    if x_point<user_path_min_x:user_path_min_x = x_point
                    if y_point<user_path_min_y:user_path_min_y = y_point
                dist+=1

            tr_x, tr_y = 0,0
            if user_path_min_x<>left_margin: tr_x = left_margin-user_path_min_x
            if user_path_min_y<>top_margin: tr_y = (top_margin-user_path_min_y)+font_size/2
            if tr_x<>0 or tr_y<>0: pdb.gimp_item_transform_translate(new_vector_id, tr_x, tr_y)
			
        else:
            line_path_width = image.width - (right_margin + left_margin) # take path length from image and user margins
            if line_path_width<(image.width/4): line_path_width = int(image.width/4) # sanity check
        size_check = True
        size_check = fn_check_widths(text, fontimg, font_size, line_path_width)
		
        if size_check == True:
		
            # add words, character widths and word widths to respective lists
            word_list, characs_widths, words_widths, total_text_width = fn2_get_width_of_words(image, fontimg, text, font_size, custom_scale, line_path_width, no_space, space_width, indent_width, kern_amount) # create lists of widths for words and characters
					
            #pdb.gimp_message("word_list =  " + str(word_list) + " : space width = " + str(space_width) + " : words_width = " + str(words_width) + " : Number of words = " + str(len(words_width)) + " : Character widths = " + str(characs_widths) + " : Number of characters = " + str(len(characs_widths) ) )
            #pdb.gimp_message("Results from fn2_get_width_of_words:")
            #pdb.gimp_message("font_size = " + str(font_size) +   " : text = " + text+ " : word_list =  " + str(word_list) + " : space width = " + str(space_width) +  " : Character widths = " + str(characs_widths) + " : Number of characters = " + str(len(characs_widths) ) + " : words_widths = " + str(words_widths) + " : total_text_width = " + str(total_text_width) )
        
		
            check_line = True
            line_loop, line_width, word_loop, temp_width, temp_line = 0, 0, 0, 0, ""
            debug_loop = 0
            temp_width = 0
            line_list = ""
            merge_check = 0
            word_list_end = len(word_list)
            next_word_text = ""
            next_word_width = 0
		
		
		    # create a list of words for each line of text
            if word_list_end>1:
                word_width = words_widths[word_loop]
                word_len = len(word_list[word_loop])
                line_width = word_width
                line_text = word_list[word_loop]
                ctr = 1
                if total_text_width>=line_path_width: 
			
                    current_line_text = word_list[word_loop]
                    current_word_width = words_widths[word_loop]
                    current_line_width = current_word_width
			
                    while word_loop<word_list_end:
                        if word_loop<word_list_end-1:
                            next_word_text = word_list[word_loop+1]
                            next_word_width = words_widths[word_loop+1]
                        else:
                            next_word_text = ""
                            next_word_width = 0
						
                        if current_line_width+space_width+next_word_width<line_path_width:
                            current_line_text+=" "+next_word_text
                            current_line_width+=space_width+next_word_width
						
                        elif current_line_width+space_width+next_word_width>=line_path_width:
                            if line_loop == 0:
                                line_list = [current_line_text]
                                line_loop+=1
                            else:
                                line_list.append(current_line_text)
                                line_loop+=1
                            current_line_text = ""
                            
                            #word_loop+=1
                            if word_loop<word_list_end-1:
                                current_line_text = word_list[word_loop+1]
                                current_word_width = words_widths[word_loop+1]
                                current_line_width = current_word_width+indent_width
							
                        word_loop += 1
                    
                        if word_loop == word_list_end: 
                            if current_line_text<>"":
                                if line_loop == 0:
                                    line_list = [current_line_text]
                                    line_loop+=1
                                else:
                                    line_list.append(current_line_text)
                                    line_loop+=1
					
					
                else:
                    word_loop, line_text = 0, ""	
                    while word_loop<len(word_list):
                        if word_loop == 0:
                            line_text+=word_list[word_loop]
                        else:
                            line_text+=" "
                            line_text+= word_list[word_loop]
                        word_loop+=1
                    line_list = [line_text]

                				
				
            else:
                word_width = words_widths[word_loop]
                word_len = len(word_list[word_loop])
                line_width = word_width
                line_text = word_list[word_loop]
                line_list = [line_text]
            # Debug messages
            #pdb.gimp_message("Final line_text = " + str(line_text) + " word word_loop = " + str(word_loop) + " line_list = " + str(line_list) )
            #pdb.gimp_message("Final line_text line_list = " + str(line_list) )
		
		
		
            # add the lines of text to a list
        
            line_loop =0
            line_x0_pos = left_margin
            line_x1_pos = image.width - right_margin		

            curr_loop, word_loop = 0, 0
            working_width = 0
            char_width_loop = 0
            current_word = ""
            max_x_pos,max_y_pos = 0, 0
		
            while line_loop<len(line_list):
            
                if line_loop == 0: # create or re-position path for characters
			
                    if user_path:
                        dummy=0
                        line_y_pos = ((font_size / 2) + top_margin)
                    else:
                        # add line path for characters
                        #line_y_pos = ((font_size / 2) + top_margin+(line_spacing*line_loop))
                        line_y_pos = ((font_size / 2) + top_margin)

                        new_vector_name = "skinny line guide path " + str((line_loop + 1))
                        new_vector_id = pdb.gimp_vectors_new(image, new_vector_name)
                        pdb.gimp_image_insert_vectors(image, new_vector_id, None, -1)
                        points1 = [line_x0_pos,line_y_pos] * 3
                        points2 = [line_x1_pos,line_y_pos] * 3
                        points = points1 + points2
                        new_vector = pdb.gimp_vectors_stroke_new_from_points(new_vector_id ,0, len(points), points, False)
                        NumberOfStrokes_in_Path,new_vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(new_vector_id)
                        new_vector_StrokeID = new_vector_StrokeIDs_List[0]
                        #pdb.gimp_item_set_visible(new_vector_id,True) # temp debug
                else:
                    if current_word <> "\n":
                        # adjust position for next line
                        #line_y_pos += ((font_size)+(line_spacing*line_loop))
                        line_y_pos += ((font_size)+(line_spacing))
                        if merge_options == 1: # reset layer merging for line merge option
                            merge_check = 0
                            if (user_path and follow_path): # re-position user line layer 
                                offset_x, offset_y = pdb.gimp_drawable_offsets(copy)
                                if copy.width+left_margin>max_x_pos:
                                    max_x_pos = copy.width+left_margin
                                if offset_x < 0:
                                    offset_x=left_margin+(abs(offset_x))
                                if offset_x > 0:
                                    offset_x=left_margin-offset_x
                                
                                #pdb.gimp_layer_translate(copy, offset_x, 0)
                                pdb.gimp_layer_translate(copy, offset_x, 0)
						
                        #pdb.gimp_item_transform_translate(new_vector_id, 0, (font_size))
                        #pdb.gimp_item_transform_translate(new_vector_id, 0, ((font_size)+(line_spacing*line_loop)))
                        pdb.gimp_item_transform_translate(new_vector_id, 0, ((font_size)+(line_spacing)))
                #pdb.gimp_message("line_loop = " + str(line_loop) + " : line_y_pos = " + str(line_y_pos) )
            
                word_loop = 0
                space_ctr = 0
                working_width = 0
                if line_loop>0:
                    if str(line_list[line_loop-1])=="\n":
                        working_width = indent_width

                current_word = line_list[line_loop]
                while word_loop<len(current_word) and current_word <> "\n":
                    if current_word[word_loop] == " ":
                        char_layer = " "
                        char_width = space_width
                    else:
                        char_layer = pdb.gimp_image_get_layer_by_name(fontimg, current_word[word_loop]+".png" ) # get name of character layer from list
                        char_width = characs_widths[char_width_loop]
                        char_width_loop += 1
				
                        copy = pdb.gimp_layer_new_from_drawable(char_layer,image) # copy and insert layer into image
                        pdb.gimp_image_insert_layer(image,copy,None,0)
                        pdb.gimp_item_set_visible(copy,True)
                        # sort out scaling to get the font size we need
                        custom_width = char_layer.width * custom_scale
                        custom_height = copy.height * custom_scale
                        old_interpolation = pdb.gimp_context_get_interpolation() # Note current interpolation method
                        pdb.gimp_context_set_interpolation(3) # Ensure that interpolation is set to Sinc (Lancos 3) method
                        pdb.gimp_layer_scale(copy,custom_width,custom_height,False)
                        pdb.gimp_context_set_interpolation(old_interpolation) # Restore old interpolation method
                        pdb.gimp_layer_set_offsets(copy,0,0)
				
                        # Get the x,y position of the required point along the active path to position the characters
                        x_point,y_point,SlopeAtThisPoint,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(new_vector_id,new_vector_StrokeID,working_width,0.01)
					
                        x_point1,y_point1,SlopeAtThisPoint1,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(new_vector_id,new_vector_StrokeID,working_width-5,0.01)
                        x_point2,y_point2,SlopeAtThisPoint2,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(new_vector_id,new_vector_StrokeID,working_width+5,0.01)
					
                        translate_x = x_point #  using returned x position 
                        if (translate_x+custom_width)>max_x_pos:max_x_pos = translate_x+custom_width # note the size for re-sizing later
							
                        translate_y = (y_point-(float(copy.height)/2))
					
                        if translate_y+(font_size/2)>max_y_pos: max_y_pos =  translate_y+(font_size/2)
						
						
                        pdb.gimp_layer_translate(copy,translate_x ,translate_y) # move layer to position along path
					
					
                        if (user_path and follow_path): # user wants custom font characters to follow the slope of the path
                            angle = computeThetaWithSlope([x_point1,y_point1], [x_point2,y_point2],SlopeAtThisPoint)
                            temp = pdb.gimp_item_transform_rotate(copy,angle,False,translate_x+float(copy.width)/2,translate_y+float(copy.height/2))
					
					
                        # Attempt to fit canvas to output
                        if y_point+(float(copy.height)/2) + font_size<>image.height: im_resize = True
                        if max_x_pos+right_margin<>image.width: im_resize = True
					
                        if (merge_options == 2 and merge_check>0): # merge for all characters
                            copy = pdb.gimp_image_merge_down(image, copy,0)
                        elif (merge_options == 1 and merge_check>0): # merge for each line of text
                            copy = pdb.gimp_image_merge_down(image, copy,0)
                        temp = pdb.gimp_displays_flush()
                        merge_check+=1
                        offset_x_adj = 0
                        if (user_path and follow_path) and merge_options==0: # reposition layer for sloping characters with no merge
                            offset_x, offset_y = pdb.gimp_drawable_offsets(copy)
                            if offset_x < 0:
                                offset_x_adj=left_margin+(abs(offset_x))
                            elif offset_x>0:
                                offset_x_adj = 0
                            if offset_y>0:
                                pdb.gimp_layer_translate(copy, 0, 0)
                                if copy.width+offset_x>max_x_pos:
                                    max_x_pos = copy.width+offset_x
                            else:
                                pdb.gimp_layer_translate(copy, offset_x+offset_x_adj, 0)
                                if copy.width+offset_x>max_x_pos:
                                    max_x_pos = copy.width+offset_x
                            if max_x_pos+right_margin<>image.width: im_resize = True
				
                    working_width += char_width # update width variable
				
               
                    word_loop+=1

                if current_word == "\n":
                    char_width_loop += 1
            
				
                line_loop+=1
				
            if (follow_path and user_path) and merge_options ==2: # merging all layers
                offset_x, offset_y = pdb.gimp_drawable_offsets(copy)
                if offset_x < 0:
                    offset_x=left_margin+(abs(offset_x))
                    if copy.width+left_margin>max_x_pos:
                        max_x_pos = copy.width+left_margin
                    if offset_y>0:
                        offset_y = top_margin-offset_y
                    pdb.gimp_layer_translate(copy, offset_x, offset_y)
                elif offset_x>0:
                    offset_x = left_margin-offset_x
                    if copy.width+left_margin>max_x_pos:
                        max_x_pos = copy.width+left_margin
                    if offset_y>0:
                        offset_y = top_margin-offset_y
                    pdb.gimp_layer_translate(copy, offset_x, offset_y)
                else:
                    offset_x = 0
                    pdb.gimp_layer_translate(copy, offset_x, offset_y)
				
			
            if merge_options == 1 and (user_path and follow_path): #re-position final line for line merge
		
                offset_x, offset_y = pdb.gimp_drawable_offsets(copy)
                if copy.width+left_margin>max_x_pos:
                    max_x_pos = copy.width+left_margin
                if offset_x < 0:
                    offset_x=left_margin+(abs(offset_x))
                if offset_x > 0:
                    offset_x=left_margin-offset_x
                pdb.gimp_layer_translate(copy, offset_x, 0)
		
            if (do_resize and im_resize): 
                pdb.gimp_image_resize(image, max_x_pos+right_margin, max_y_pos + (font_size/2) + bottom_margin, 0, 0)
                    
                
            if not user_path: pdb.gimp_image_remove_vectors(image, new_vector_id)
		
            if user_path and remove_user_path: pdb.gimp_image_remove_vectors(image, new_vector_id)
			
            
            # get layers information
            num_layers, layer_ids = pdb.gimp_image_get_layers(image)
            layer_list = image.layers
            loop, layer_ctr = 0, 0
            while loop<num_layers:
                layer_name = pdb.gimp_item_get_name(layer_list[loop])
                if ".png" not in layer_name: 
                    pdb.gimp_layer_resize_to_image_size(layer_list[loop])
                    if layer_ctr == 0: bg_layer = layer_list[loop]
                    layer_ctr+=1
                loop+=1
				
            if background_type>0:
                # sort out colours for fill etc
                original_fg_col = pdb.gimp_context_get_foreground()
                original_bg_col = pdb.gimp_context_get_background()
                fg_col, bg_col = original_fg_col, original_bg_col
                if col_style==0: # reverse colours
                    fg_col, bg_col = original_bg_col, original_fg_col
                    pdb.gimp_context_set_foreground(fg_col)
                    pdb.gimp_context_set_background(bg_col)
					
                if background_type==1: # plain fill
                    pdb.gimp_drawable_fill(bg_layer, 0)
		
                elif background_type==2: # gradient fill
                    pdb.gimp_context_set_gradient_fg_bg_rgb()
                    if gradient_style == 0: # radial gradient
                        pdb.gimp_drawable_edit_gradient_fill(bg_layer, 2, 0, False, 1, 0, True, image.width/2, image.height/2, image.width, image.height)
                    else:
                        pdb.gimp_drawable_edit_gradient_fill(bg_layer, 0, 0, False, 1, 0, True, image.width/2, 0, image.width/2, image.height)
						
                elif background_type==3: # feathered selection
                    pdb.gimp_drawable_fill(bg_layer, 1) # fill background colour
                    layer_pos = pdb.gimp_image_get_item_position(image, bg_layer)
                    feather_layer = pdb.gimp_layer_new(image, image.width, image.height, 1, "feather layer", 100, 28)
                    pdb.gimp_image_insert_layer(image, feather_layer, None, layer_pos)
                    pdb.gimp_image_select_item(image, 2, bg_layer)
                    pdb.gimp_selection_shrink(image, selection_width)
                    pdb.gimp_selection_feather(image, feather_val)
                    pdb.gimp_drawable_edit_fill(feather_layer, 0) # fill foreground colour
                    pdb.gimp_selection_none(image)
                    bg_layer = pdb.gimp_image_merge_down(image, feather_layer, 2)
                        
            if drop_colour>0:
				
                num_layers, layer_ids = pdb.gimp_image_get_layers(image)
                layer_list = image.layers
                loop, layer_ctr = 0, 0
                while loop<num_layers:
                    this_layer = layer_list[loop]
                    layer_name = pdb.gimp_item_get_name(this_layer)
                    if ".png" in layer_name: 
                        this_layer = proc_dropshadow(image, this_layer, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour, background_type)
                    loop+=1
				
                    
                       
                
            pdb.gimp_displays_flush()
			
			
        else:
		
            message_text = "The current font size is too large.\nPlease reduce the size and try again."
            old_handler = pdb.gimp_message_get_handler()	
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message(message_text)
            pdb.gimp_message_set_handler(old_handler)    
        
    else:
        message_text = "Please select a custom font and try again."
        old_handler = pdb.gimp_message_get_handler()	
        pdb.gimp_message_set_handler(MESSAGE_BOX)
        pdb.gimp_message(message_text)
        pdb.gimp_message_set_handler(old_handler)    
	
    
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
######################################################################################################################################

	
def fn2_get_width_of_words(image, fontimg, text, font_size, custom_scale, line_path_width, no_space, space_width, indent_width, kern_amount): # returns widths of words and characters as lists

    curr_loop, word_width, prev_word_pos, word_loop, total_text_width, line_width = 0, 0, 0, 0, 0, 0
    temp_word, size_check = "", True
    new_line=False
    while curr_loop < len(text):
        current_char = text[curr_loop]
        if current_char==" " or current_char=="\n":
    	    if current_char == " ":
                if word_loop == 0:
                    temp_word = text[prev_word_pos:curr_loop]
                    if temp_word<>"": word_list = [text[prev_word_pos:curr_loop]]
                    if len(text[prev_word_pos:curr_loop])==1: 
                        if temp_word<>"": words_widths = [kerning_width]
                    else:
                        if temp_word<>"": words_widths = [word_width]
                    total_text_width = words_widths[0]
                    word_loop += 1
					
                    word_width = 0
					
                else:
                    temp_word = text[prev_word_pos:curr_loop]
                    if temp_word<>"": word_list.append( text[prev_word_pos:curr_loop])
                    if len(text[prev_word_pos:curr_loop])==1: 
                        if temp_word<>"": 
                            words_widths.append(kerning_width)
                            total_text_width += (space_width+kerning_width)
                    else:
                        if temp_word<>"": 
                            words_widths.append(word_width)
                            total_text_width += (space_width+word_width)
                    word_loop += 1
                    word_width = 0
                prev_word_pos = curr_loop+1
            else: # deal with newline character
                new_line = True
                temp_word = text[prev_word_pos:curr_loop]
                indent_char_pos = curr_loop+1
                if word_loop == 0:
                    if temp_word<>"": word_list = [text[prev_word_pos:curr_loop]]
                    if temp_word<>"": 
                        words_widths = [word_width]
                        total_text_width = word_width
                    word_loop += 1
                    if temp_word<>"": 
                        words_widths.append(line_path_width)
                        total_text_width += line_path_width
                    word_width = line_path_width
					
                else:
                    word_list.append( text[prev_word_pos:curr_loop])
                    if temp_word<>"": 
                        words_widths.append(word_width)
                        total_text_width += (space_width+word_width)
                    word_loop += 1
                    if temp_word<>"": 
                        words_widths.append(line_path_width)
                        total_text_width += line_path_width
                    word_width = line_path_width
					
                prev_word_pos = curr_loop+1
				
                if curr_loop == 0:
                    characs_widths = [line_path_width]
                else:
                    characs_widths.append(line_path_width)
                    word_list.append(text[curr_loop]) # adds newline character to word_list
            '''
            '''
        else:
            current_width, kerning_width = 0, 0
            if curr_loop < len(text)-1: 
                next_char = text[curr_loop+1]
            else:
                next_char = ""
        
            # check for available 2 character kerning paths 
            kerning_name = current_char + next_char + ".png"
            kerning_vector = pdb.gimp_image_get_vectors_by_name(fontimg,kerning_name) # check for character kerning pair paths with .png
			
            if kerning_vector<0: # if 2 character kerning path doesn't exist check for single character kerning path
                kerning_name = current_char + ".png"
                kerning_vector = pdb.gimp_image_get_vectors_by_name(fontimg,kerning_name)
				
            if kerning_vector>=0: # kerning vector exists so update value of kerning_width
                # Get path stroke information
                NumberOfStrokes_in_Path,kerning_vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(kerning_vector)
                kerning_vector_StrokeID = kerning_vector_StrokeIDs_List[0]
                kerning_width = pdb.gimp_vectors_stroke_get_length(kerning_vector,kerning_vector_StrokeID,0.01)
                kerning_width = (kerning_width*custom_scale)+kern_amount
			
            if kerning_width==0:
                current_layer_name = current_char + ".png"
                char_layer = pdb.gimp_image_get_layer_by_name(fontimg, current_layer_name)
				
                kerning_width = ((char_layer.width)*custom_scale)+kern_amount
				
            word_width += kerning_width
			
            if word_width>=line_path_width:
                temp_word = text[prev_word_pos:curr_loop]
                if word_loop == 0:
                    word_width -= kerning_width
                    if temp_word<>"": word_list = [text[prev_word_pos:curr_loop]]
                    if temp_word<>"": 
                        words_widths = [word_width]
                        total_text_width = word_width
                    word_loop += 1
                    word_width = 0
                    word_width = kerning_width
                else:
                    
                    if temp_word<>"": word_list.append( text[prev_word_pos:curr_loop])
                    word_width -= kerning_width
                    if temp_word<>"": 
                        words_widths.append(word_width)
                        total_text_width +=  (space_width+word_width)
                    word_loop += 1
                    word_width = 0
                    word_width = kerning_width
                prev_word_pos = curr_loop
			
			
            if curr_loop == 0:
                characs_widths = [kerning_width]
            else:
                characs_widths.append(kerning_width)
					
        curr_loop+=1
		
    temp_word = text[prev_word_pos:curr_loop]
    if word_loop==0:
        if temp_word<>"": word_list=[text[prev_word_pos:len(text)]]
        words_widths=[word_width]
        total_text_width = word_width
    else:
        if temp_word<>"": 
            word_list.append( text[prev_word_pos:len(text)])
        words_widths.append(word_width)
        total_text_width +=  (space_width+word_width)
    return word_list, characs_widths, words_widths, total_text_width
	
	
# ---------------------------------------------------------------------------------

def computeThetaWithSlope(fromPoint,toPoint,slope):
    dX=toPoint[0]-fromPoint[0]
    dY=toPoint[1]-fromPoint[1]

    if math.fabs(dX) < .001:
        theta=math.copysign(math.pi/2,dY)
    else:
        theta=math.atan(slope)
        if dX<0:
            theta=theta+math.pi
    return theta	
	
#######################################################################################################

def fn_check_widths(text, fontimg, f_size, line_path_width):
    loop, new_text, test = 0,"", True
    while loop<len(text):
        char = text[loop]
        if char<>"\n":
            layer_name = char + ".png"
            char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
            if char <> " ":
                temp_width = char_layer.width
                temp_width = temp_width*fn_get_custom_font_scale(f_size, char_layer.height)
                if temp_width>line_path_width: test = False
        loop += 1
    return test


def fn_check_text_is_valid(text, fontimg):
    # checks that characters are present in the custom font, replacing any missing characters with a space
    loop, new_text, test = 0,"", True
    while loop<len(text):
        char = text[loop]
        #if char == "'": pdb.gimp_message("Apostrophe detected")
        
        if char<>"\n":
            layer_name = char + ".png"
            #pdb.gimp_message(layer_name)
            char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
            #pdb.gimp_message(layer_name + " : " + str(char_layer) )
            if char_layer < 0:
                if char.lower() == char:
                    char = char.upper()
                else:
                    char = char.lower()
                layer_name = char + ".png"
                char_layer = pdb.gimp_image_get_layer_by_name(fontimg,layer_name)
                if char_layer < 0:
                    char = " "
            
       
        new_text += char
           
        loop += 1
		
    return new_text


#######################################################################################################

def fn_get_custom_font_scale(f_size, custom_font_height):
    new_custom_height = f_size * 1.0
    scale = new_custom_height / custom_font_height
    return scale

#######################################################################################################

def fn_test_valid_custom_font(xcffilename):
    test = False
    fontimg = -1
    if xcffilename > "":
        try:
            fontimg = pdb.gimp_xcf_load(0,xcffilename,xcffilename)
        except RuntimeError:
            fontimg = -1
		
        else:
            fontimg = pdb.gimp_xcf_load(0,xcffilename,xcffilename)
            fontimg_layer = pdb.gimp_image_get_layer_by_name(fontimg, "space.png")
            if fontimg_layer >= 0: test = True
    return test, fontimg

# -------------------------------------------------------------------------------------

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	
######################################################################################################################################


######################################################################################################################################

 	
#####################################################################################################

def proc_dropshadow(im, target_layer, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour, background_type):
    #REM Create a text drop shadow effect - optionally clipped to the text layer for transparent type texts
	
    # make a copy of the image width/height
    old_im_width, old_im_height = im.width, im.height
    name = pdb.gimp_item_get_name(target_layer)
    # Copy layer to use as the drop shadow
    layer_pos = pdb.gimp_image_get_item_position(im, target_layer)
    shadow_layer = pdb.gimp_layer_copy(target_layer, False)
    # Add and re-position the layer in the layer stack
    im.add_layer(shadow_layer, layer_pos)
    #if background_type<>3: pdb.gimp_image_lower_item(im, shadow_layer)
    pdb.gimp_image_lower_item(im, shadow_layer)
    # get layer offsets for the shadow layer (which will also be the same as the original layer holding the user text)
    shadow_offset_x, shadow_offset_y = pdb.gimp_drawable_offsets(shadow_layer)
    # work out the centre point of the shadow/original layer
    shad_centre_x = (shadow_layer.width/2)
    shad_centre_y = (shadow_layer.height/2)
    # make up new size for the shadow layer to accommodate blurred result - currently twice the existing width and height
    new_shad_width = shadow_layer.width*2
    new_shad_height = shadow_layer.height*2
    # work out new offsets for the new size of the shadow layer to ensure that the shadow and original layers are still aligned
    new_shad_layer_offset_x = 0 - ((shadow_layer.width - new_shad_width)/2)
    new_shad_layer_offset_y = 0 - ((shadow_layer.height - new_shad_height)/2)
    # and apply the new offsets to the shadow layer
    pdb.gimp_layer_resize(shadow_layer, new_shad_width, new_shad_height, new_shad_layer_offset_x, new_shad_layer_offset_y)
    # make a note of the new shadow layer offsets
    shadow_resized_offset_x, shadow_resized_offset_y = pdb.gimp_drawable_offsets(shadow_layer)	
    # if any of the shadow layer falls outside the image boundaries
    reset_shadow = False
    '''
    if new_shad_layer_offset_x<0 or new_shad_layer_offset_y<0 or shadow_layer.width > im.width or shadow_layer.height > im.height:
        reset_shadow = True
        # set offsets of the shadow layer to image origin
        pdb.gimp_layer_set_offsets(shadow_layer,0,0)
        # set offsets for original text layer to match
        pdb.gimp_layer_set_offsets(target_layer,0 + abs(new_shad_layer_offset_x) ,0 + abs(new_shad_layer_offset_y))
        # resize the image to accomodate the new shadow layer size
        new_im_width, new_im_height = im.width, im.height
        if shadow_layer.width > im.width: new_im_width = shadow_layer.width
        if shadow_layer.height > im.height: new_im_height = shadow_layer.height
        pdb.gimp_image_resize(im, new_im_width, new_im_height, 0, 0)
    '''
    # get alpha selection
    pdb.gimp_image_select_item(im, 2, shadow_layer)
    # fill the text on the shadow layer with the background colour
    if drop_colour<3:
        pdb.gimp_drawable_edit_fill(shadow_layer,drop_colour-1)
    if drop_colour==4:
        pdb.gimp_context_set_foreground(black)
        pdb.gimp_drawable_edit_fill(shadow_layer,0)
    # deselect
    pdb.gimp_selection_none(im)
    # blur shadow 
    pdb.plug_in_gauss(im, shadow_layer, drop_blur, drop_blur, 1)
    # crop the blurred layer
    pdb.plug_in_autocrop_layer(im, shadow_layer)
    # move the shadow layer to user position and adjust opacity
    temp = pdb.gimp_layer_translate(shadow_layer, drop_offset_x, drop_offset_y)
    pdb.gimp_layer_set_opacity(shadow_layer, drop_opacity)
    # merge and rename
    target_layer = pdb.gimp_image_merge_down(im,target_layer,0)
    pdb.gimp_item_set_name(target_layer, name)
    if reset_shadow: 
        cropped_offset_x, cropped_offset_y = pdb.gimp_drawable_offsets(target_layer)	
        pdb.gimp_layer_set_offsets(target_layer,shadow_offset_x, shadow_offset_y)
    pdb.gimp_image_resize(im, old_im_width, old_im_height, 0, 0)
    

    #### Other Useful Functions
# -------------------------------------------------------------------------------------
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height


######################################################################################################################################
	
def fn_selection_amount(sel_value,r_ness,adjustment):
    # 
    value = (math.sqrt(sel_value)*r_ness)*adjustment
    return value

######################################################################################################################################
	
def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)
	
######################################################################################################################################
######################################################################################################################################
	
register(
    "python_fu_Custom_Font_Paragrapher_v1_0_Tabbed",
    "Pat's simple paragraph writer for use with Custom Fonts",
    "Pat's simple paragraphs writer for use with Custom Fonts",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2023",
    "<Image>/Filters/Custom Fonts/Using Tools/Custom Font Paragrapher v1.0 Tabbed...",
    "RGB*",
    [
    (PF_FILENAME, "pf_names_xcffilename", ".xcf custom font file: ", 0),
    (PF_TEXT, "pf_string", "Text: ", "Gimp is fun."),
    (PF_SLIDER, "pf_fontsize", "Font size: ", 100, (8, 600, 1)),
    (PF_SLIDER, "pf_linespace", "Line spacing amount: ", 0, (-100, 500, 1)),
    #(PF_TOGGLE, "spacing_start", "Apply line spacing adjustment: ", 1, [("After 1st line"),("After 2nd line") ]),
    (PF_SLIDER, "pf_Charspace", "Character spacing amount: ", 0, (-100, 200, 1)),
    (PF_OPTION, "pf_indent_option", "Indent options", 0, ("No indents", "Indent new paragraphs")),
    (PF_SLIDER, "pf_indent_size", "Number of spaces widths for indent: ", 2, (1, 8, 1)),
	
    (PF_SLIDER, "pf_top", "Top margin amount: ", 50, (0, 250, 1),{'newLine':'1','tab':1}),
    (PF_SLIDER, "pf_left", "Left margin amount: ", 50, (0, 250, 1),{'newLine':'1','tab':1}),
    (PF_SLIDER, "pf_right", "Right margin amount: ", 50, (0, 250, 1),{'newLine':'1','tab':1}),
    (PF_SLIDER, "pf_bottom", "Bottom margin amount: ", 50, (0, 250, 1),{'newLine':'1','tab':1}),
    (PF_OPTION, "pf_merge_option", "Merge layers options", 2, ("Individual characters", "Merge characters into lines", "Merge all characters"),{'newLine':'1','tab':1}),
    (PF_TOGGLE, "copy_paths", "Use active path as line template\n(if available): ", 0, [("Text along lines"),("Text along user path") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "follow_path", "Rotate text to follow path slope: ", 0, [("Keep text vertical"),("Rotate text to slope") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "remove_path", "Remove user path after using: ", 1, [("Keep"),("Remove") ],{'newLine':'1','tab':1}),
	
    (PF_OPTION, "bg_option", "Background options: ", 0, ("None", "Use a plain fill", "Use a gradient fill", "Use a feathered selection"),{'newLine':'1','tab':2}),
    (PF_TOGGLE, "fg_bg", "Fill style: ", 1, [("Use BG colour"),("Use FG colour") ],{'newLine':'1','tab':2}),
    (PF_TOGGLE, "gradient", "Use radial or linear gradient: ", 1, [("Radial Gradient"),("Linear Gradient") ],{'newLine':'1','tab':2}),
    (PF_SLIDER, "pf_sel_width", "Selection width: ", 50, (0, 500, 10),{'newLine':'1','tab':2}),
    (PF_SLIDER, "pf_feather", "Feather amount: ", 200, (0, 500, 10),{'newLine':'1','tab':2}),
    (PF_TOGGLE, "re_sizer", "Resize image to fit text: ", 1, [("Do not re-size image"),("Re-size image") ],{'newLine':'1','tab':2}),
	
	
    (PF_OPTION, "drop_colour", "Drop shadow", 0, ( "No Drop Shadow", "Use Foreground Colour", "Use Background colour", "Use a Layer copy", "Use Black"),{'newLine':'1','tab':3}),
    (PF_SLIDER, "pf_drop_offset_x", "Shadow offset x position: ", 8, (-50, 50, 1),{'newLine':'1','tab':3}),
    (PF_SLIDER, "pf_drop_offset_y", "Shadow offset y position: ", 8, (-50, 50, 1),{'newLine':'1','tab':3}),
    (PF_SLIDER, "pf_drop_blur", "Shadow blur percentage: ", 15, (0, 50, 1),{'newLine':'1','tab':3}),
    (PF_SLIDER, "pf_drop_opacity", "Shadow opacity: ", 70, (20, 100, 1),{'newLine':'1','tab':3}),
	
    (PF_TAB    , "ptab"    , "tabs", 0, (_('Font/Spacing Options'),_('Path/Merge/Margin Options'),_('Background options'),_('Drop Shadow options')) ),
    ],
    [],
    do_stuff)

main()
